# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/coveragepy/coveragepy/blob/main/NOTICE.txt

"""Tests for concurrency libraries."""

from __future__ import annotations

import glob
import multiprocessing
import os
import pathlib
import random
import re
import sys
import threading
import time

from types import ModuleType
from collections.abc import Iterable

import pytest

import coverage
from coverage import env
from coverage.data import line_counts
from coverage.exceptions import ConfigError
from coverage.files import abs_file
from coverage.misc import import_local_file
from coverage.sqldata import SUFFIX_PATTERN

from tests import testenv
from tests.coveragetest import CoverageTest


# These libraries aren't always available, we'll skip tests if they aren't.

try:
    import gevent
except ImportError:
    gevent = None

try:
    import greenlet
except ImportError:
    greenlet = None


def measurable_line(l: str) -> bool:
    """Is this a line of code coverage will measure?

    Not blank, not a comment, and not "else"
    """
    l = l.strip()
    if not l:
        return False
    if l.startswith("#"):
        return False
    if l.startswith("else:"):
        return False
    return True


def line_count(s: str) -> int:
    """How many measurable lines are in `s`?"""
    return len(list(filter(measurable_line, s.splitlines())))


def print_simple_annotation(code: str, linenos: Iterable[int]) -> None:
    """Print the lines in `code` with X for each line number in `linenos`."""
    for lineno, line in enumerate(code.splitlines(), start=1):
        print(" {} {}".format("X" if lineno in linenos else " ", line))


class LineCountTest(CoverageTest):
    """Test the helpers here."""

    run_in_temp_dir = False

    def test_line_count(self) -> None:
        CODE = """
            # Hey there!
            x = 1
            if x:
                print("hello")
            else:
                print("bye")

            print("done")
            """

        assert line_count(CODE) == 5


# The code common to all the concurrency models.
SUM_RANGE_Q = """
    # Above this will be imports defining queue and threading.

    class Producer(threading.Thread):
        def __init__(self, limit, q):
            threading.Thread.__init__(self)
            self.limit = limit
            self.q = q

        def run(self):
            for i in range(self.limit):
                self.q.put(i)
            self.q.put(None)

    class Consumer(threading.Thread):
        def __init__(self, q, qresult):
            threading.Thread.__init__(self)
            self.q = q
            self.qresult = qresult

        def run(self):
            sum = 0
            while "no peephole".upper():
                i = self.q.get()
                if i is None:
                    break
                sum += i
            self.qresult.put(sum)

    def sum_range(limit):
        q = queue.Queue()
        qresult = queue.Queue()
        c = Consumer(q, qresult)
        p = Producer(limit, q)
        c.start()
        p.start()

        p.join()
        c.join()
        return qresult.get()

    # Below this will be something using sum_range.
    """

PRINT_SUM_RANGE = """
    print(sum_range({QLIMIT}))
    """

# Import the things to use threads.
THREAD = """
    import threading
    import queue
    """

# Import the things to use gevent.
GEVENT = """
    from gevent import monkey
    monkey.patch_thread()
    import threading
    import gevent.queue as queue
    """

# Uncomplicated code that doesn't use any of the concurrency stuff, to test
# the simple case under each of the regimes.
SIMPLE = """
    total = 0
    for i in range({QLIMIT}):
        total += i
    print(total)
    """


def cant_trace_msg(concurrency: str, the_module: ModuleType | None) -> str | None:
    """What might coverage.py say about a concurrency setting and imported module?"""
    # In the concurrency choices, "multiprocessing" doesn't count.
    assert "multiprocessing" not in concurrency

    if testenv.SYS_MON and concurrency:
        expected_out = f"Can't use core=sysmon: it doesn't support concurrency={concurrency}"
    elif the_module is None:
        # We don't even have the underlying module installed, we expect
        # coverage to alert us to this fact.
        expected_out = f"Couldn't trace with concurrency={concurrency}, the module isn't installed."
    elif testenv.C_TRACER or concurrency == "thread" or concurrency == "":
        expected_out = None
    else:
        expected_out = (
            f"Can't support concurrency={concurrency} with {testenv.TRACER_CLASS}, "
            + "only threads are supported."
        )
    return expected_out


class ConcurrencyTest(CoverageTest):
    """Tests of the concurrency support in coverage.py."""

    QLIMIT = 1000

    def try_some_code(
        self,
        code: str,
        concurrency: str,
        the_module: ModuleType,
        expected_out: str | None = None,
    ) -> None:
        """Run some concurrency testing code and see that it was all covered.

        `code` is the Python code to execute.  `concurrency` is the name of
        the concurrency regime to test it under.  `the_module` is the imported
        module that must be available for this to work at all. `expected_out`
        is the text we expect the code to produce.

        """
        self.make_file("try_it.py", code)

        cmd = f"coverage run --concurrency={concurrency} try_it.py"
        _, out = self.run_command_status(cmd)

        expected_cant_trace = cant_trace_msg(concurrency, the_module)

        if expected_cant_trace is not None:
            assert expected_cant_trace in out
            pytest.skip(f"Can't test: {expected_cant_trace}")
        else:
            # We can fully measure the code if we are using the C tracer, which
            # can support all the concurrency, or if we are using threads.
            if expected_out is None:
                expected_out = "%d\n" % (sum(range(self.QLIMIT)))
            print(code)
            assert out == expected_out

            # Read the coverage file and see that try_it.py has all its lines
            # executed.
            data = coverage.CoverageData(".coverage")
            data.read()

            # If the test fails, it's helpful to see this info:
            fname = abs_file("try_it.py")
            linenos = data.lines(fname)
            assert linenos is not None
            print(f"{len(linenos)}: {linenos}")
            print_simple_annotation(code, linenos)

            lines = line_count(code)
            assert line_counts(data)["try_it.py"] == lines

    @pytest.mark.skipif(
        not testenv.CAN_MEASURE_THREADS, reason="Can't measure threads with this core."
    )
    def test_threads(self) -> None:
        code = (THREAD + SUM_RANGE_Q + PRINT_SUM_RANGE).format(QLIMIT=self.QLIMIT)
        self.try_some_code(code, "thread", threading)

    @pytest.mark.skipif(
        not testenv.CAN_MEASURE_THREADS, reason="Can't measure threads with this core."
    )
    def test_threads_simple_code(self) -> None:
        code = SIMPLE.format(QLIMIT=self.QLIMIT)
        self.try_some_code(code, "thread", threading)

    # https://github.com/coveragepy/coveragepy/issues/663
    @pytest.mark.skipif(env.WINDOWS, reason="gevent has problems on Windows: #663")
    def test_gevent(self) -> None:
        code = (GEVENT + SUM_RANGE_Q + PRINT_SUM_RANGE).format(QLIMIT=self.QLIMIT)
        self.try_some_code(code, "gevent", gevent)

    def test_gevent_simple_code(self) -> None:
        code = SIMPLE.format(QLIMIT=self.QLIMIT)
        self.try_some_code(code, "gevent", gevent)

    def test_greenlet(self) -> None:
        GREENLET = """\
            from greenlet import greenlet

            def test1(x, y):
                z = gr2.switch(x+y)
                print(z)

            def test2(u):
                print(u)
                gr1.switch(42)

            gr1 = greenlet(test1)
            gr2 = greenlet(test2)
            gr1.switch("hello", " world")
            """
        self.try_some_code(GREENLET, "greenlet", greenlet, "hello world\n42\n")

    def test_greenlet_simple_code(self) -> None:
        code = SIMPLE.format(QLIMIT=self.QLIMIT)
        self.try_some_code(code, "greenlet", greenlet)

    # The code in tracer.c that went with this test doesn't seem particular to
    # eventlet. I don't want to remove this test, but I don't know how to
    # rewrite it to demonstrate the original problem without eventlet.
    @pytest.mark.skip(reason="We don't test eventlet; don't know how to rewrite this test.")
    def test_bug_330(self) -> None:
        BUG_330 = """\
            from weakref import WeakKeyDictionary
            import eventlet

            def do():
                eventlet.sleep(.01)

            gts = WeakKeyDictionary()
            for _ in range(100):
                gts[eventlet.spawn(do)] = True
                eventlet.sleep(.005)

            eventlet.sleep(.1)
            print(len(gts))
            """
        eventlet = glob  # quiet linters on the next line.
        self.try_some_code(BUG_330, "eventlet", eventlet, "0\n")

    # Sometimes a test fails due to inherent randomness. Try more times.
    @pytest.mark.skipif(
        not testenv.CAN_MEASURE_THREADS, reason="Can't measure threads with this core."
    )
    @pytest.mark.flaky(max_runs=3)
    def test_threads_with_gevent(self) -> None:
        self.make_file(
            "both.py",
            """\
            import queue
            import threading

            import gevent

            def work1(q):
                q.put(1)

            def gwork(q):
                gevent.spawn(work1, q).join()
                q.put(None)
                print("done")

            q = queue.Queue()
            t = threading.Thread(target=gwork, args=(q,))
            t.start()
            t.join()

            answer = q.get()
            assert answer == 1
            """,
        )
        _, out = self.run_command_status("coverage run --concurrency=thread,gevent both.py")
        if gevent is None:
            assert "Couldn't trace with concurrency=gevent, the module isn't installed.\n" in out
            pytest.skip("Can't run test without gevent installed.")
        if not testenv.C_TRACER:
            assert testenv.PY_TRACER
            assert out == (
                "Can't support concurrency=gevent with PyTracer, only threads are supported.\n"
            )
            pytest.skip(f"Can't run gevent with {testenv.TRACER_CLASS}.")

        assert out == "done\n"

        out = self.run_command("coverage report -m")
        last_line = self.squeezed_lines(out)[-1]
        assert re.search(r"TOTAL \d+ 0 100%", last_line)

    def test_bad_concurrency(self) -> None:
        with pytest.raises(ConfigError, match="Unknown concurrency choices: nothing"):
            self.command_line("run --concurrency=nothing prog.py")

    def test_bad_concurrency_in_config(self) -> None:
        self.make_file(".coveragerc", "[run]\nconcurrency = nothing\n")
        with pytest.raises(ConfigError, match="Unknown concurrency choices: nothing"):
            self.command_line("run prog.py")

    def test_no_multiple_light_concurrency(self) -> None:
        with pytest.raises(ConfigError, match="Conflicting concurrency settings: gevent, greenlet"):
            self.command_line("run --concurrency=gevent,greenlet prog.py")

    def test_no_multiple_light_concurrency_in_config(self) -> None:
        self.make_file(".coveragerc", "[run]\nconcurrency = gevent, greenlet\n")
        with pytest.raises(ConfigError, match="Conflicting concurrency settings: gevent, greenlet"):
            self.command_line("run prog.py")

    def test_multiprocessing_needs_config_file(self) -> None:
        with pytest.raises(ConfigError, match="multiprocessing requires a configuration file"):
            self.command_line("run --concurrency=multiprocessing prog.py")


class WithoutConcurrencyModuleTest(CoverageTest):
    """Tests of what happens if the requested concurrency isn't installed."""

    @pytest.mark.parametrize("module", ["gevent", "greenlet"])
    def test_missing_module(self, module: str) -> None:
        self.make_file("prog.py", "a = 1")
        sys.modules[module] = None  # type: ignore[assignment]
        msg = rf"Couldn't trace with concurrency={module}, the module isn't installed."
        with pytest.raises(ConfigError, match=msg):
            self.command_line(f"run --concurrency={module} prog.py")


SQUARE_OR_CUBE_WORK = """
    def work(x):
        # Use different lines in different subprocesses.
        if x % 2:
            y = x*x
        else:
            y = x*x*x
        return y
    """

SUM_RANGE_WORK = """
    def work(x):
        return sum_range((x+1)*100)
    """

MULTI_CODE = """
    # Above this will be a definition of work().
    import multiprocessing
    import os
    import time
    import sys

    def process_worker_main(args):
        # Need to pause, or the tasks go too quickly, and some processes
        # in the pool don't get any work, and then don't record data.
        ret = work(*args)
        time.sleep(0.1)
        return os.getpid(), ret

    if __name__ == "__main__":      # pragma: no branch
        # This if is on a single line so we can get 100% coverage
        # even if we have no arguments.
        if len(sys.argv) > 1: multiprocessing.set_start_method(sys.argv[1])
        pool = multiprocessing.Pool({NPROCS})
        inputs = [(x,) for x in range({UPTO})]
        outputs = pool.imap_unordered(process_worker_main, inputs)
        pids = set()
        total = 0
        for pid, sq in outputs:
            pids.add(pid)
            total += sq
        print(f"{{len(pids)}} pids, {{total = }}")
        pool.close()
        pool.join()
    """


@pytest.fixture(params=["fork", "spawn"], name="start_method")
def start_method_fixture(request: pytest.FixtureRequest) -> str:
    """Parameterized fixture to choose the start_method for multiprocessing."""
    start_method: str = request.param
    if start_method not in multiprocessing.get_all_start_methods():
        # Windows doesn't support "fork".
        pytest.skip(f"start_method={start_method} not supported here")
    return start_method


# Sometimes a test fails due to inherent randomness. Try more times.
# @pytest.mark.flaky(max_runs=30)
class MultiprocessingTest(CoverageTest):
    """Test support of the multiprocessing module."""

    def try_multiprocessing_code(
        self,
        code: str,
        expected_out: str | None,
        nprocs: int,
        start_method: str,
        concurrency: str = "multiprocessing",
        args: str = "",
    ) -> None:
        """Run code using multiprocessing, it should produce `expected_out`."""
        self.make_file("multi.py", code)
        self.make_file(
            ".coveragerc",
            f"""\
            [run]
            concurrency = {concurrency}
            source = .
            """,
        )

        cmd = f"coverage run {args} multi.py {start_method}"
        _, out = self.run_command_status(cmd)
        assert out.rstrip() == expected_out
        assert len(glob.glob(".coverage.*")) == nprocs + 1

        out = self.run_command("coverage combine --debug=combine")
        out_lines = out.splitlines()
        # main proc log file, nprocs log files, one summary line
        assert len(out_lines) == 1 + nprocs + 1
        assert all(
            re.fullmatch(
                rf"(Combined data file|Skipping duplicate data) \.coverage{SUFFIX_PATTERN}",
                line,
            )
            for line in out_lines[:-1]
        )
        # Combines the main and one subprocess, the other subprocesses are skipped.
        assert out_lines[-1] == f"Combined 2 files, skipped {nprocs - 1}"
        assert len(glob.glob(".coverage.*")) == 0
        out = self.run_command("coverage report -m")

        last_line = self.squeezed_lines(out)[-1]
        assert re.search(r"TOTAL \d+ 0 100%", last_line)

    def test_multiprocessing_simple(self, start_method: str) -> None:
        nprocs = 3
        upto = 30
        code = (SQUARE_OR_CUBE_WORK + MULTI_CODE).format(NPROCS=nprocs, UPTO=upto)
        total = sum(x * x if x % 2 else x * x * x for x in range(upto))
        expected_out = f"{nprocs} pids, {total = }"
        self.try_multiprocessing_code(
            code,
            expected_out,
            nprocs,
            start_method=start_method,
        )

    def test_multiprocessing_append(self, start_method: str) -> None:
        nprocs = 3
        upto = 30
        code = (SQUARE_OR_CUBE_WORK + MULTI_CODE).format(NPROCS=nprocs, UPTO=upto)
        total = sum(x * x if x % 2 else x * x * x for x in range(upto))
        expected_out = f"{nprocs} pids, total = {total}"
        self.try_multiprocessing_code(
            code,
            expected_out,
            nprocs,
            args="--append",
            start_method=start_method,
        )

    @pytest.mark.skipif(
        not testenv.CAN_MEASURE_BRANCHES, reason="Can't measure branches with this core"
    )
    def test_multiprocessing_with_branching(self, start_method: str) -> None:
        nprocs = 3
        upto = 30
        code = (SQUARE_OR_CUBE_WORK + MULTI_CODE).format(NPROCS=nprocs, UPTO=upto)
        total = sum(x * x if x % 2 else x * x * x for x in range(upto))
        expected_out = f"{nprocs} pids, total = {total}"
        self.make_file("multi.py", code)
        self.make_file(
            "multi.rc",
            """\
            [run]
            concurrency = multiprocessing
            branch = True
            omit = */site-packages/*
            """,
        )

        out = self.run_command(f"coverage run --rcfile=multi.rc multi.py {start_method}")
        assert out.rstrip() == expected_out

        out = self.run_command("coverage combine -q")  # sneak in a test of -q
        assert out == ""
        out = self.run_command("coverage report -m")

        last_line = self.squeezed_lines(out)[-1]
        assert re.search(r"TOTAL \d+ 0 \d+ 0 100%", last_line)

    def test_multiprocessing_bootstrap_error_handling(self) -> None:
        # An exception during bootstrapping will be reported.
        self.make_file(
            "multi.py",
            """\
            import multiprocessing
            if __name__ == "__main__":
                with multiprocessing.Manager():
                    pass
            """,
        )
        self.make_file(
            ".coveragerc",
            """\
            [run]
            concurrency = multiprocessing
            _crash = _bootstrap
            """,
        )
        out = self.run_command("coverage run multi.py", status=1)
        assert "Exception during multiprocessing bootstrap init" in out
        assert "RuntimeError: Crashing because called by _bootstrap" in out

    def test_bug_890(self) -> None:
        # chdir in multiprocessing shouldn't keep us from finding the
        # .coveragerc file.
        self.make_file(
            "multi.py",
            """\
            import multiprocessing, os, os.path
            if __name__ == "__main__":
                if not os.path.exists("./tmp"): os.mkdir("./tmp")
                os.chdir("./tmp")
                with multiprocessing.Manager():
                    pass
                print("ok")
            """,
        )
        self.make_file(
            ".coveragerc",
            """\
            [run]
            concurrency = multiprocessing
            """,
        )
        out = self.run_command("coverage run multi.py")
        assert out.splitlines()[-1] == "ok"


@pytest.mark.skipif(not testenv.SETTRACE_CORE, reason="gettrace is not supported with this core.")
def test_coverage_stop_in_threads() -> None:
    has_started_coverage = []
    has_stopped_coverage = []

    def run_thread() -> None:  # pragma: nested
        """Check that coverage is stopping properly in threads."""
        deadline = time.time() + 5
        ident = threading.current_thread().ident
        if sys.gettrace() is not None:
            has_started_coverage.append(ident)
        while sys.gettrace() is not None:
            # Wait for coverage to stop
            time.sleep(0.01)
            if time.time() > deadline:
                return
        has_stopped_coverage.append(ident)

    cov = coverage.Coverage()
    with cov.collect():
        t = threading.Thread(target=run_thread)
        t.start()

        time.sleep(0.1)
    t.join()

    assert has_started_coverage == [t.ident]
    assert has_stopped_coverage == [t.ident]


def test_thread_safe_save_data(tmp_path: pathlib.Path) -> None:
    # Non-regression test for: https://github.com/coveragepy/coveragepy/issues/581

    # Create some Python modules and put them in the path
    modules_dir = tmp_path / "test_modules"
    modules_dir.mkdir()
    module_names = [f"m{i:03d}" for i in range(1000)]
    for module_name in module_names:
        (modules_dir / (module_name + ".py")).write_text("def f(): pass\n", encoding="utf-8")

    # Shared variables for threads
    should_run = [True]
    imported = []

    old_dir = os.getcwd()
    os.chdir(modules_dir)
    try:
        # Make sure that all dummy modules can be imported.
        for module_name in module_names:
            import_local_file(module_name)

        def random_load() -> None:  # pragma: nested
            """Import modules randomly to stress coverage."""
            while should_run[0]:
                module_name = random.choice(module_names)
                mod = import_local_file(module_name)
                mod.f()
                imported.append(mod)

        # Spawn some threads with coverage enabled and attempt to read the
        # results right after stopping coverage collection with the threads
        #  still running.
        duration = 0.01
        for _ in range(3):
            cov = coverage.Coverage()
            with cov.collect():
                threads = [threading.Thread(target=random_load) for _ in range(10)]
                should_run[0] = True
                for t in threads:
                    t.start()

                time.sleep(duration)

            # The following call used to crash with running background threads.
            cov.get_data()

            # Stop the threads
            should_run[0] = False
            for t in threads:
                t.join()

            if (not imported) and duration < 10:  # pragma: only failure
                duration *= 2

    finally:
        os.chdir(old_dir)
        should_run[0] = False


@pytest.mark.skipif(env.WINDOWS, reason="SIGTERM doesn't work the same on Windows")
@pytest.mark.flaky(max_runs=3)  # Sometimes a test fails due to inherent randomness. Try more times.
class SigtermTest(CoverageTest):
    """Tests of our handling of SIGTERM."""

    @pytest.mark.parametrize("sigterm", [False, True])
    def test_sigterm_multiprocessing_saves_data(self, sigterm: bool) -> None:
        # A terminated process should save its coverage data.
        self.make_file(
            "clobbered.py",
            """\
            import multiprocessing
            import time

            def subproc(x):
                if x.value == 3:
                    print("THREE", flush=True)  # line 6, missed
                else:
                    print("NOT THREE", flush=True)
                x.value = 0
                time.sleep(60)

            if __name__ == "__main__":
                print("START", flush=True)
                x = multiprocessing.Value("L", 1)
                proc = multiprocessing.Process(target=subproc, args=(x,))
                proc.start()
                while x.value != 0:
                    time.sleep(.05)
                proc.terminate()
                print("END", flush=True)
            """,
        )
        self.make_file(
            ".coveragerc",
            """\
            [run]
            parallel = True
            concurrency = multiprocessing
            """
            + ("sigterm = true" if sigterm else ""),
        )
        out = self.run_command("coverage run clobbered.py")
        # Under Linux, things go wrong. Does that matter?
        if env.LINUX and "assert self._collectors" in out:
            lines = out.splitlines(True)
            out = "".join(lines[:3])
        assert out == "START\nNOT THREE\nEND\n"
        self.run_command("coverage combine")
        out = self.run_command("coverage report -m")
        if sigterm:
            expected = "clobbered.py 17 1 94% 6"
        else:
            expected = "clobbered.py 17 5 71% 5-10"
        assert self.squeezed_lines(out)[2] == expected

    def test_sigterm_threading_saves_data(self) -> None:
        # A terminated process should save its coverage data.
        self.make_file(
            "handler.py",
            """\
            import os, signal

            print("START", flush=True)
            print("SIGTERM", flush=True)
            os.kill(os.getpid(), signal.SIGTERM)
            print("NOT HERE", flush=True)
            """,
        )
        self.make_file(
            ".coveragerc",
            """\
            [run]
            # The default concurrency option.
            concurrency = thread
            sigterm = true
            """,
        )
        status, out = self.run_command_status("coverage run handler.py")
        assert status != 0
        out_lines = out.splitlines()
        assert len(out_lines) in [2, 3]
        assert out_lines[:2] == ["START", "SIGTERM"]
        if len(out_lines) == 3:
            assert out_lines[2] == "Terminated"
        out = self.run_command("coverage report -m")
        expected = "handler.py 5 1 80% 6"
        assert self.squeezed_lines(out)[2] == expected

    def test_sigterm_still_runs(self) -> None:
        # A terminated process still runs its own SIGTERM handler.
        self.make_file(
            "handler.py",
            """\
            import multiprocessing
            import signal
            import time

            def subproc(x):
                print("START", flush=True)
                def on_sigterm(signum, frame):
                    print("SIGTERM", flush=True)

                signal.signal(signal.SIGTERM, on_sigterm)
                x.value = 0
                try:
                    time.sleep(.1)
                except OSError: # This happens on PyPy3.11 on Mac
                    pass
                print("END", flush=True)

            if __name__ == "__main__":
                x = multiprocessing.Value("L", 1)
                proc = multiprocessing.Process(target=subproc, args=(x,))
                proc.start()
                while x.value != 0:
                    time.sleep(.02)
                proc.terminate()
            """,
        )
        self.make_file(
            ".coveragerc",
            """\
            [run]
            parallel = True
            concurrency = multiprocessing
            sigterm = True
            """,
        )
        out = self.run_command("coverage run handler.py")
        assert out == "START\nSIGTERM\nEND\n"



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/coveragepy/coveragepy/blob/main/NOTICE.txt

"""Bytecode analysis for coverage.py"""

from __future__ import annotations

import dis
from collections.abc import Iterable, Mapping
from types import CodeType

from coverage.types import TArc, TLineNo, TOffset


class ByteParser:
    """Parse bytecode to understand the structure of code."""

    def __init__(
        self,
        *,
        code: CodeType | None = None,
        text: str | None = None,
        filename: str | None = None,
    ) -> None:
        if code is None:
            assert text is not None
            code = compile(text, filename or "<string>", "exec", dont_inherit=True)
        self.code = code

    def _child_parsers(self) -> Iterable[ByteParser]:
        """Iterate over all the code objects nested within this one.

        The iteration includes `self` as its first value.

        We skip code objects named `__annotate__` since they are deferred
        annotations that usually are never run.  If there are errors in the
        annotations, they will be caught by type checkers or other tools that
        use annotations.

        """
        return (ByteParser(code=c) for c in self.code_objects() if c.co_name != "__annotate__")

    def code_objects(self) -> Iterable[CodeType]:
        """Iterate over all the code objects in `code`."""
        stack = [self.code]
        while stack:
            # We're going to return the code object on the stack, but first
            # push its children for later returning.
            code = stack.pop()
            for c in code.co_consts:
                if isinstance(c, CodeType):
                    stack.append(c)
            yield code

    def _line_numbers(self) -> Iterable[TLineNo]:
        """Yield the line numbers possible in this code object.

        Uses co_lines() to produce a sequence: l0, l1, ...
        """
        for _, _, line in self.code.co_lines():
            if line:
                yield line

    def find_statements(self) -> Iterable[TLineNo]:
        """Find the statements in `self.code`.

        Produce a sequence of line numbers that start statements.  Recurses
        into all code objects reachable from `self.code`.

        """
        for bp in self._child_parsers():
            # Get all of the lineno information from this code.
            yield from bp._line_numbers()


def bytes_to_lines(code: CodeType) -> dict[TOffset, TLineNo]:
    """Make a dict mapping byte code offsets to line numbers."""
    b2l = {}
    for bstart, bend, lineno in code.co_lines():
        if lineno is not None:
            for boffset in range(bstart, bend, 2):
                b2l[boffset] = lineno
    return b2l


def op_set(*op_names: str) -> set[int]:
    """Make a set of opcodes from instruction names.

    The names might not exist in this version of Python, skip those if not.
    """
    ops = {op for name in op_names if (op := dis.opmap.get(name))}
    assert ops, f"At least one opcode must exist: {op_names}"
    return ops


# Opcodes that are unconditional jumps elsewhere.
ALWAYS_JUMPS = op_set(
    "JUMP_BACKWARD",
    "JUMP_BACKWARD_NO_INTERRUPT",
    "JUMP_FORWARD",
)

# Opcodes that exit from a function.
RETURNS = op_set(
    "RETURN_VALUE",
    "RETURN_GENERATOR",
)


# CACHE doesn't exist in Python 3.10, but the branch resolver is only used
# on 3.14+, so a placeholder value is fine.
_CACHE = dis.opmap.get("CACHE", -1)
_EXTENDED_ARG = dis.opmap["EXTENDED_ARG"]

# All opcodes with a jump target.
JUMPS = set(dis.hasjrel) | set(dis.hasjabs)

# Opcodes that jump backwards.
BACKWARD_JUMPS = {op for op in JUMPS if "JUMP_BACKWARD" in dis.opname[op]}


class BranchArcResolver:
    """Resolve branch events to line arcs, one (source, dest) pair at a time.

    Branch events are one-shot (they are DISABLEd after firing), so each
    (source offset, destination offset) pair is resolved at most a couple of
    times per code object.  Resolving pairs on demand is much cheaper than
    precomputing trails for every branch in the code object, most of which
    never fire.  We walk the raw bytecode bytes so that we never need to
    disassemble whole code objects with `dis`.

    To resolve one pair:
    starting from the destination, follow the trail of instructions (through
    unconditional jumps) until we reach an instruction on a new source line
    (giving us the arc), a return (an arc to leaving the code object), or
    another branch possibility (no arc: that branch will produce its own
    events).

    """

    def __init__(
        self,
        code: CodeType,
        byte_to_line: Mapping[TOffset, TLineNo],
        multiline_map: Mapping[TLineNo, TLineNo],
    ) -> None:
        self.code = code
        # co_code re-copies the bytes on each access, so fetch it once.
        self.co_code = code.co_code
        self.byte_to_line = byte_to_line
        self.multiline_map = multiline_map

    def line_at(self, offset: TOffset) -> TLineNo | None:
        """The source line of the instruction at `offset`, de-multilined."""
        line = self.byte_to_line.get(offset)
        if line is not None:
            line = self.multiline_map.get(line, line)
        return line

    def resolve(self, source: TOffset, dest: TOffset) -> TArc | None:
        """Turn a branch event's (source, dest) offsets into an arc, or None."""
        from_line = self.line_at(source)
        if from_line is None:
            return None
        co_code = self.co_code
        max_offset = len(co_code)
        byte_to_line = self.byte_to_line
        multiline_map = self.multiline_map
        offset = dest
        ext_arg = 0
        seen: set[TOffset] = set()
        while 0 <= offset < max_offset and offset not in seen:
            seen.add(offset)
            op = co_code[offset]
            if op == _CACHE:
                offset += 2
                continue
            if op == _EXTENDED_ARG:
                ext_arg = (ext_arg | co_code[offset + 1]) << 8
                offset += 2
                continue
            line = byte_to_line.get(offset)
            if line is not None:
                line = multiline_map.get(line, line)
                if line and line != from_line:
                    return (from_line, line)
            if op in JUMPS:
                if op in ALWAYS_JUMPS:
                    arg = ext_arg | co_code[offset + 1]
                    # Jump distances are measured from the end of the
                    # instruction's inline CACHE entries, which appear in
                    # co_code as CACHE opcodes immediately following it.
                    next_offset = offset + 2
                    while next_offset < max_offset and co_code[next_offset] == _CACHE:
                        next_offset += 2
                    if op in BACKWARD_JUMPS:
                        offset = next_offset - 2 * arg
                    else:
                        offset = next_offset + 2 * arg
                    ext_arg = 0
                    continue
                # Another branch possibility: it will get its own events.
                return None
            if op in RETURNS:
                return (from_line, -self.code.co_firstlineno)
            ext_arg = 0
            offset += 2
        return None



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/coveragepy/coveragepy/blob/main/NOTICE.txt

"""Hypothesis strategies for use in tests."""

from __future__ import annotations

from collections.abc import Sequence

from hypothesis import strategies as st

# Type aliases:
# S: a Hypothesis strategy, generic.
# P: primitive leaf data types.
# D: nested data.

S = st.SearchStrategy
P = None | bool | int | float | str | bytes
D = P | list["D"] | tuple["D", ...] | set["D"] | dict[str, "D"]

# Leaf types, all hashable. Some are limited to the kinds of values coverage
# deals with, to focus Hypothesis.
primitives: Sequence[S[P]] = [
    st.none(),
    st.booleans(),
    st.integers(min_value=-1000, max_value=100_000),
    st.floats(min_value=-100, max_value=100, allow_nan=False, allow_infinity=False).map(
        # Avoid -0.0.
        lambda x: 0.0 if x == 0.0 else x
    ),
    st.text(max_size=10),
    st.binary(max_size=10),
]


def homogenous_tuples_of(elements: S[D]) -> S[tuple[D, ...]]:
    """Make a strategy for homogenous tuples.

    The elements are all the same type, one of the strategies from `elements`.
    The tuples will be varying lengths.
    """
    return st.lists(elements, max_size=3).map(tuple)


def true_tuple_strategies(element_strategies: Sequence[S[D]]) -> S[S[tuple[D, ...]]]:
    """Make a strategy that produces tuple-generating strategies.

    The elements will be types drawn from `element_strategies`.
    The tuples drawn from the returned strategy will have heterogenous types,
    but every tuple will have the same types in the same positions, and all the
    tuples will be the same length.
    """
    # Note the way we use this, we won't ever get nested tuples: (int, (str, int))
    return st.lists(st.sampled_from(element_strategies), max_size=3).map(
        lambda strats: st.tuples(*strats)
    )


# A strategy for producing strategies that produce hashable values.
hashable_strategies: S[S[D]] = st.one_of(
    st.sampled_from(primitives),
    true_tuple_strategies(primitives),
)


def dicts_of(leaves: S[D]) -> S[dict[str, D]]:
    """Make a strategy for dicts with string keys and values from `leaves`."""
    return st.dictionaries(st.text(max_size=10), leaves, max_size=3)


def nested_dicts_of(leaves: S[D]) -> S[dict[str, D]]:
    """Make a strategy for recursive dicts with leaves from another strategy."""
    return dicts_of(st.recursive(leaves, dicts_of, max_leaves=5))


# A strategy that generates strategies for nested data.
nested_data_strategies: S[S[D]] = st.recursive(
    hashable_strategies,
    lambda children: st.one_of(
        children.map(lambda s: st.lists(s, max_size=5)),
        children.map(homogenous_tuples_of),
        hashable_strategies.map(lambda s: st.sets(s, max_size=5)),
        children.map(nested_dicts_of),
    ),
    max_leaves=3,
)

if __name__ == "__main__":
    for _ in range(100):
        print(repr(nested_data_strategies.example().example()))



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/coveragepy/coveragepy/blob/main/NOTICE.txt

"""Run a simple trace function on a file of Python code."""

import os, sys

nest = 0


def trace(frame, event, arg):
    global nest

    if nest is None:
        # This can happen when Python is shutting down.
        return None

    if the_program in frame.f_code.co_filename:
        print(
            "%s%s %s %d @%d"
            % (
                "    " * nest,
                event,
                os.path.basename(frame.f_code.co_filename),
                frame.f_lineno,
                frame.f_lasti,
            )
        )

    if event == "call":
        nest += 1
    if event == "return":
        nest -= 1

    return trace


print(sys.version)
the_program = sys.argv[1]

code = open(the_program, encoding="utf-8").read()
code_obj = compile(code, the_program, mode="exec")
sys.settrace(trace)
exec(code_obj)



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/coveragepy/coveragepy/blob/main/NOTICE.txt

"""Tests for coverage.py's improved tokenizer."""

from __future__ import annotations

import os.path
import re
import sys
import textwrap
import warnings

import pytest

from coverage import env
from coverage.phystokens import source_token_lines, source_encoding
from coverage.python import get_python_source

from tests.coveragetest import CoverageTest, TESTS_DIR


# A simple program and its token stream.
SIMPLE = """\
# yay!
def foo():
  say('two = %d' % 2)
"""

SIMPLE_TOKENS = [
    [("com", "# yay!")],
    [("key", "def"), ("ws", " "), ("nam", "foo"), ("op", "("), ("op", ")"), ("op", ":")],
    [
        ("ws", "  "),
        ("nam", "say"),
        ("op", "("),
        ("str", "'two = %d'"),
        ("ws", " "),
        ("op", "%"),
        ("ws", " "),
        ("num", "2"),
        ("op", ")"),
    ],
]

# Mixed-white-space program, and its token stream.
MIXED_WS = """\
def hello():
        a="Hello world!"
\tb="indented"
"""

MIXED_WS_TOKENS = [
    [("key", "def"), ("ws", " "), ("nam", "hello"), ("op", "("), ("op", ")"), ("op", ":")],
    [("ws", "        "), ("nam", "a"), ("op", "="), ("str", '"Hello world!"')],
    [("ws", "        "), ("nam", "b"), ("op", "="), ("str", '"indented"')],
]

# https://github.com/coveragepy/coveragepy/issues/822
BUG_822 = """\
print( "Message 1" )
array = [ 1,2,3,4,       # 4 numbers \\
          5,6,7 ]        # 3 numbers
print( "Message 2" )
"""


class PhysTokensTest(CoverageTest):
    """Tests for coverage.py's improved tokenizer."""

    run_in_temp_dir = False

    def check_tokenization(self, source: str) -> None:
        """Tokenize `source`, then put it back together, should be the same."""
        tokenized = ""
        for line in source_token_lines(source):
            text = "".join(t for _, t in line)
            tokenized += text + "\n"
        # source_token_lines doesn't preserve trailing spaces, so trim all that
        # before comparing.
        source = source.replace("\r\n", "\n")
        source = re.sub(r"(?m)[ \t]+$", "", source)
        tokenized = re.sub(r"(?m)[ \t]+$", "", tokenized)
        assert source == tokenized

    def check_file_tokenization(self, fname: str) -> None:
        """Use the contents of `fname` for `check_tokenization`."""
        self.check_tokenization(get_python_source(fname))

    def test_simple(self) -> None:
        assert list(source_token_lines(SIMPLE)) == SIMPLE_TOKENS
        self.check_tokenization(SIMPLE)

    def test_missing_final_newline(self) -> None:
        # We can tokenize source that is missing the final newline.
        assert list(source_token_lines(SIMPLE.rstrip())) == SIMPLE_TOKENS

    def test_tab_indentation(self) -> None:
        # Mixed tabs and spaces...
        assert list(source_token_lines(MIXED_WS)) == MIXED_WS_TOKENS

    def test_bug_822(self) -> None:
        self.check_tokenization(BUG_822)

    def test_tokenize_real_file(self) -> None:
        # Check the tokenization of a real file (large, btw).
        real_file = os.path.join(TESTS_DIR, "test_coverage.py")
        self.check_file_tokenization(real_file)

    def test_1828(self) -> None:
        # https://github.com/coveragepy/coveragepy/pull/1828
        tokens = list(
            source_token_lines(
                textwrap.dedent("""
                    x = \\
                        1
                    a = ["aaa",\\
                         "bbb \\
                         ccc"]
                """)
            )
        )
        assert tokens == [
            [],
            [("nam", "x"), ("ws", " "), ("op", "="), ("ws", " "), ("xx", "\\")],
            [("ws", "    "), ("num", "1")],
            [
                ("nam", "a"),
                ("ws", " "),
                ("op", "="),
                ("ws", " "),
                ("op", "["),
                ("str", '"aaa"'),
                ("op", ","),
                ("xx", "\\"),
            ],
            [("ws", "     "), ("str", '"bbb \\')],
            [("str", '     ccc"'), ("op", "]")],
        ]

    @pytest.mark.parametrize(
        "fname",
        [
            "stress_phystoken.tok",
            "stress_phystoken_dos.tok",
        ],
    )
    def test_stress(self, fname: str) -> None:
        # Check the tokenization of the stress-test files.
        # And check that those files haven't been incorrectly "fixed".
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message=r".*invalid escape sequence")

            stress = os.path.join(TESTS_DIR, fname)
            self.check_file_tokenization(stress)
            with open(stress, encoding="utf-8") as fstress:
                assert re.search(r"(?m) $", fstress.read()), f"{stress} needs a trailing space."

    def test_fstring_middle(self) -> None:
        tokens = list(
            source_token_lines(
                textwrap.dedent("""\
                    f'Look: {x} {{x}}!'
                """)
            )
        )
        if env.PYBEHAVIOR.fstring_syntax:
            assert tokens == [
                [
                    ("fst", "f'"),
                    ("fst", "Look: "),
                    ("op", "{"),
                    ("nam", "x"),
                    ("op", "}"),
                    ("fst", " {{"),
                    ("fst", "x}}"),
                    ("fst", "!"),
                    ("fst", "'"),
                ],
            ]
        else:
            assert tokens == [[("str", "f'Look: {x} {{x}}!'")]]


class SoftKeywordTest(CoverageTest):
    """Tests the tokenizer handling soft keywords."""

    run_in_temp_dir = False

    def test_soft_keywords_match_case(self) -> None:
        source = textwrap.dedent("""\
            match re.match(something):
                case ["what"]:
                    match = case("hello")
                case [_]:
                    match("hello")
                    match another.thing:
                        case 1:
                            pass

            class case(): pass
            def match():
                global case
            """)
        tokens = list(source_token_lines(source))
        assert tokens[0][0] == ("key", "match")
        assert tokens[0][4] == ("nam", "match")
        assert tokens[1][1] == ("key", "case")
        assert tokens[2][1] == ("nam", "match")
        assert tokens[2][5] == ("nam", "case")
        assert tokens[3][1] == ("key", "case")
        assert tokens[4][1] == ("nam", "match")
        assert tokens[5][1] == ("key", "match")
        assert tokens[6][1] == ("key", "case")
        assert tokens[9][2] == ("nam", "case")
        assert tokens[10][2] == ("nam", "match")
        assert tokens[11][3] == ("nam", "case")

    @pytest.mark.skipif(sys.version_info < (3, 12), reason="type isn't a soft keyword until 3.12")
    def test_soft_keyword_type(self) -> None:
        source = textwrap.dedent("""\
            type Point = tuple[float, float]
            type(int)
            """)
        tokens = list(source_token_lines(source))
        assert tokens[0][0] == ("key", "type")
        assert tokens[1][0] == ("nam", "type")

    @pytest.mark.skipif(sys.version_info < (3, 15), reason="lazy isn't a soft keyword until 3.15")
    def test_soft_keyword_lazy(self) -> None:
        source = textwrap.dedent("""\
            lazy import foo
            lazy from foo import bar
            def lazy(f): pass
            lazy(12)
            """)
        tokens = list(source_token_lines(source))
        assert tokens[0][0] == ("key", "lazy")
        assert tokens[1][0] == ("key", "lazy")
        assert tokens[2][2] == ("nam", "lazy")
        assert tokens[3][0] == ("nam", "lazy")


# The default source file encoding.
DEF_ENCODING = "utf-8"


ENCODING_DECLARATION_SOURCES = [
    # Various forms from http://www.python.org/dev/peps/pep-0263/
    (1, b"# coding=cp850\n\n", "cp850"),
    (1, b"# coding=latin-1\n", "iso-8859-1"),
    (1, b"# coding=iso-latin-1\n", "iso-8859-1"),
    (1, b"#!/usr/bin/python\n# -*- coding: cp850 -*-\n", "cp850"),
    (1, b"#!/usr/bin/python\n# vim: set fileencoding=cp850:\n", "cp850"),
    (1, b"# This Python file uses this encoding: cp850\n", "cp850"),
    (1, b"# This file uses a different encoding:\n# coding: cp850\n", "cp850"),
    (1, b"\n# coding=cp850\n\n", "cp850"),
    (2, b"# -*-  coding:cp850 -*-\n# vim: fileencoding=cp850\n", "cp850"),
]


class SourceEncodingTest(CoverageTest):
    """Tests of source_encoding() for detecting encodings."""

    run_in_temp_dir = False

    def test_detect_source_encoding(self) -> None:
        for _, source, expected in ENCODING_DECLARATION_SOURCES:
            assert source_encoding(source) == expected, f"Wrong encoding in {source!r}"

    def test_detect_source_encoding_not_in_comment(self) -> None:
        # Should not detect anything here
        source = b"def parse(src, encoding=None):\n    pass"
        assert source_encoding(source) == DEF_ENCODING

    def test_dont_detect_source_encoding_on_third_line(self) -> None:
        # A coding declaration doesn't count on the third line.
        source = b"\n\n# coding=cp850\n\n"
        assert source_encoding(source) == DEF_ENCODING

    def test_detect_source_encoding_of_empty_file(self) -> None:
        # An important edge case.
        assert source_encoding(b"") == DEF_ENCODING

    def test_bom(self) -> None:
        # A BOM means utf-8.
        source = b"\xef\xbb\xbftext = 'hello'\n"
        assert source_encoding(source) == "utf-8-sig"

    def test_bom_with_encoding(self) -> None:
        source = b"\xef\xbb\xbf# coding: utf-8\ntext = 'hello'\n"
        assert source_encoding(source) == "utf-8-sig"

    def test_bom_is_wrong(self) -> None:
        # A BOM with an explicit non-utf8 encoding is an error.
        source = b"\xef\xbb\xbf# coding: cp850\n"
        with pytest.raises(SyntaxError, match="encoding problem: utf-8"):
            source_encoding(source)

    def test_unknown_encoding(self) -> None:
        source = b"# coding: klingon\n"
        with pytest.raises(SyntaxError, match="unknown encoding: klingon"):
            source_encoding(source)



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/coveragepy/coveragepy/blob/main/NOTICE.txt

"""Code coverage measurement for Python"""

# Setuptools setup for coverage.py
# This file is used unchanged under all versions of Python.

import re
import os
import os.path
import platform
import sys
import textwrap
import zipfile

from pathlib import Path
from typing import Any

from setuptools import Extension, errors, setup
from setuptools.command.build_ext import build_ext
from setuptools.command.editable_wheel import editable_wheel


def get_version_data() -> dict[str, Any]:
    """Get the globals from coverage/version.py."""
    # We exec coverage/version.py so we can avoid importing the product code into setup.py.
    module_globals: dict[str, Any] = {}
    cov_ver_py = os.path.join(os.path.split(__file__)[0], "coverage/version.py")
    with open(cov_ver_py, encoding="utf-8") as version_file:
        # Execute the code in version.py.
        exec(compile(version_file.read(), cov_ver_py, "exec", dont_inherit=True), module_globals)
    return module_globals


def get_long_description(url: str) -> str:
    """Massage README.rst to get the long description."""
    with open("README.rst", encoding="utf-8") as readme:
        readme_text = readme.read()

    url = url.replace("readthedocs", "@@")
    assert "@@" not in readme_text
    long_description = (
        readme_text.replace("https://coverage.readthedocs.io/en/latest", url)
        .replace("https://coverage.readthedocs.io", url)
        .replace("@@", "readthedocs")
    )
    return long_description


def count_contributors() -> int:
    """Read CONTRIBUTORS.txt to count how many people have helped."""
    with open("CONTRIBUTORS.txt", "rb") as contributors:
        paras = contributors.read().split(b"\n\n")
        num_others = len(paras[-1].splitlines())
        num_others += 1  # Count Gareth Rees, who is mentioned in the top paragraph.
    return num_others


# PYVERSIONS
CLASSIFIERS = textwrap.dedent("""\
    Development Status :: 5 - Production/Stable
    Environment :: Console
    Intended Audience :: Developers
    Operating System :: OS Independent
    Programming Language :: Python
    Programming Language :: Python :: 3
    Programming Language :: Python :: 3.10
    Programming Language :: Python :: 3.11
    Programming Language :: Python :: 3.12
    Programming Language :: Python :: 3.13
    Programming Language :: Python :: 3.14
    Programming Language :: Python :: 3.15
    Programming Language :: Python :: 3.16
    Programming Language :: Python :: Free Threading :: 3 - Stable
    Programming Language :: Python :: Implementation :: CPython
    Programming Language :: Python :: Implementation :: PyPy
    Topic :: Software Development :: Quality Assurance
    Topic :: Software Development :: Testing
""").splitlines()


# The names of .pth files matter because they are read in lexicographic order.
# Editable installs work because of `__editable__*.pth` files, so we need our
# .pth files to come after those. But we want ours to be earlyish in the
# sequence, so we start with `a`. The metacov .pth file should come before the
# coverage .pth file, so we use `a0_metacov.pth` and `a1_coverage.pth`.
PTH_NAME = "a1_coverage.pth"


def make_pth_file() -> None:
    """Make the packaged .pth file used for measuring subprocess coverage."""

    with open("coverage/pth_file.py", encoding="utf-8") as f:
        code = f.read()

    code = re.sub(r"\s*#.*\n", "\n", code)
    code = code.replace("    ", " ").strip()

    # `import sys` is needed because .pth files are executed only if they start
    # with `import `.
    with open(PTH_NAME, "w", encoding="utf-8") as pth_file:
        pth_file.write(f"import sys; exec({code!r})\n")


class EditableWheelWithPth(editable_wheel):  # type: ignore[misc]
    """Override the editable_wheel command to insert our .pth file into the wheel."""

    def run(self) -> None:
        super().run()
        for whl in Path(self.dist_dir).glob("*editable*.whl"):
            with zipfile.ZipFile(whl, "a") as zf:
                zf.write(PTH_NAME, PTH_NAME)


# A replacement for the build_ext command which raises a single exception
# if the build fails, so we can fallback nicely.

ext_errors = (
    errors.CCompilerError,
    errors.ExecError,
    errors.PlatformError,
)
if sys.platform == "win32":
    # IOError can be raised when failing to find the compiler
    ext_errors += (IOError,)


class BuildFailed(Exception):
    """Raise this to indicate the C extension wouldn't build."""

    def __init__(self) -> None:
        Exception.__init__(self)
        self.cause = sys.exc_info()[1]  # work around py 2/3 different syntax


class ve_build_ext(build_ext):  # type: ignore[misc]
    """Build C extensions, but fail with a straightforward exception."""

    def run(self) -> None:
        """Wrap `run` with `BuildFailed`."""
        try:
            build_ext.run(self)
        except errors.PlatformError as exc:
            raise BuildFailed() from exc

    def build_extension(self, ext: Any) -> None:
        """Wrap `build_extension` with `BuildFailed`."""
        if self.compiler.compiler_type == "msvc":
            ext.extra_compile_args = (ext.extra_compile_args or []) + [
                "/std:c11",
                "/experimental:c11atomics",
            ]
        try:
            # Uncomment to test compile failure handling:
            #   raise errors.CCompilerError("OOPS")
            build_ext.build_extension(self, ext)
        except ext_errors as exc:
            raise BuildFailed() from exc
        except ValueError as err:
            # this can happen on Windows 64 bit, see Python issue 7511
            if "'path'" in str(err):  # works with both py 2/3
                raise BuildFailed() from err
            raise


version_data = get_version_data()
make_pth_file()


# Create the keyword arguments for setup()

setup_args = dict(
    name="coverage",
    version=version_data["__version__"],
    packages=[
        "coverage",
    ],
    package_data={
        "coverage": [
            "htmlfiles/*.*",
            "py.typed",
            f"../{PTH_NAME}",
        ],
    },
    entry_points={
        "console_scripts": [
            # Install a script as "coverage".
            "coverage = coverage.cmdline:main",
            # And as "coverage3", and as "coverage-3.7" (or whatever), but deprecated.
            "coverage%d = coverage.cmdline:main_deprecated" % sys.version_info[:1],
            "coverage-%d.%d = coverage.cmdline:main_deprecated" % sys.version_info[:2],
        ],
    },
    extras_require={
        # Enable pyproject.toml support.
        "toml": ['tomli; python_full_version<="3.11.0a6"'],
    },
    cmdclass={
        "build_ext": ve_build_ext,
        "editable_wheel": EditableWheelWithPth,
    },
    # We need to get HTML assets from our htmlfiles directory.
    zip_safe=False,
    author=f"Ned Batchelder and {count_contributors()} others",
    author_email="ned@nedbatchelder.com",
    description=__doc__,
    long_description=get_long_description(url=version_data["__url__"]),
    long_description_content_type="text/x-rst",
    keywords="code coverage testing",
    license="Apache-2.0",
    license_files=["LICENSE.txt"],
    classifiers=CLASSIFIERS,
    url="https://github.com/coveragepy/coveragepy",
    project_urls={
        "Documentation": version_data["__url__"],
        "Funding": (
            "https://tidelift.com/subscription/pkg/pypi-coverage"
            + "?utm_source=pypi-coverage&utm_medium=referral&utm_campaign=pypi"
        ),
        "Issues": "https://github.com/coveragepy/coveragepy/issues",
        "Mastodon": "https://hachyderm.io/@coveragepy",
        "Mastodon (nedbat)": "https://hachyderm.io/@nedbat",
    },
    python_requires=">=3.10",  # minimum of PYVERSIONS
)

# There are a few reasons we might not be able to compile the C extension.
# Figure out if we should attempt the C extension or not. Define
# COVERAGE_DISABLE_EXTENSION in the build environment to explicitly disable the
# extension.

compile_extension = os.getenv("COVERAGE_DISABLE_EXTENSION", None) is None

if platform.python_implementation() == "PyPy":
    # Pypy can't compile C extensions
    compile_extension = False

if compile_extension:
    setup_args.update(
        dict(
            ext_modules=[
                Extension(
                    "coverage.tracer",
                    sources=[
                        "coverage/ctracer/datastack.c",
                        "coverage/ctracer/filedisp.c",
                        "coverage/ctracer/module.c",
                        "coverage/ctracer/tracer.c",
                    ],
                ),
            ],
        ),
    )


def main() -> None:
    """Actually invoke setup() with the arguments we built above."""
    # For a variety of reasons, it might not be possible to install the C
    # extension.  Try it with, and if it fails, try it without.
    try:
        setup(**setup_args)
    except BuildFailed as exc:
        msg = "Couldn't install with extension module, trying without it..."
        exc_msg = f"{exc.__class__.__name__}: {exc.cause}"
        print(f"**\n** {msg}\n** {exc_msg}\n**")

        del setup_args["ext_modules"]
        setup(**setup_args)


if __name__ == "__main__":
    main()
